@@ -1,5 +1,3 @@
-#!/usr/bin/env python
-# -*- coding: utf-8 -*-
 #
 # Copyright 2004-2006 Zuza Software Foundation
 #
@@ -19,16 +17,16 @@
 # along with this program; if not, see <http://www.gnu.org/licenses/>.
 
 """Handles converting of files between formats (used by
-:mod:`translate.convert` tools)."""
+:mod:`translate.convert` tools).
+"""
 
 import os.path
-try:
-    from cStringIO import StringIO
-except ImportError:
-    from StringIO import StringIO
+from io import BytesIO
 
 from translate.misc import optrecurse
-# don't import optparse ourselves, get the version from optrecurse
+
+
+# Don't import optparse ourselves, get the version from optrecurse.
 optparse = optrecurse.optparse
 
 
@@ -38,9 +36,9 @@ class ConvertOptionParser(optrecurse.RecursiveOptionParser, object):
     def __init__(self, formats, usetemplates=False, usepots=False,
                  allowmissingtemplate=False, description=None):
         """construct the specialized Option Parser"""
-        optrecurse.RecursiveOptionParser.__init__(self, formats, usetemplates,
-                                                  allowmissingtemplate=allowmissingtemplate,
-                                                  description=description)
+        super().__init__(formats, usetemplates,
+                         allowmissingtemplate=allowmissingtemplate,
+                         description=description)
         self.usepots = usepots
         self.settimestampoption()
         self.setpotoption()
@@ -60,6 +58,15 @@ class ConvertOptionParser(optrecurse.RecursiveOptionParser, object):
                         action="store_false", default=default, help=nofuzzyhelp)
         self.passthrough.append("includefuzzy")
 
+    def add_remove_untranslated_option(self, default=False):
+        """Adds an option to remove key value from output if it is
+        untranslated.
+        """
+        self.add_option("", "--removeuntranslated", dest="remove_untranslated",
+                        default=False, action="store_true",
+                        help="remove untranslated strings from output")
+        self.passthrough.append("remove_untranslated")
+
     def add_threshold_option(self, default=None):
         """Adds an option to output only stores where translation percentage
         exceeds the threshold.
@@ -74,7 +81,7 @@ class ConvertOptionParser(optrecurse.RecursiveOptionParser, object):
         self.add_option(
             "", "--duplicates", dest="duplicatestyle", default=default,
             type="choice", choices=["msgctxt", "merge"],
-            help="what to do with duplicate strings (identical source text): merge, msgctxt (default: '%s')" % \
+            help="what to do with duplicate strings (identical source text): merge, msgctxt (default: '%s')" %
                  default,
             metavar="DUPLICATESTYLE"
         )
@@ -111,7 +118,7 @@ class ConvertOptionParser(optrecurse.RecursiveOptionParser, object):
             potformat = self.potifyformat(fileformat)
             if potformat != fileformat:
                 helpformats.append(potformat)
-        return super(ConvertOptionParser, self).getformathelp(helpformats)
+        return super().getformathelp(helpformats)
 
     def filterinputformats(self, options):
         """Filters input formats, processing relevant switches in options."""
@@ -124,7 +131,7 @@ class ConvertOptionParser(optrecurse.RecursiveOptionParser, object):
         """Filters output options, processing relevant switches in options."""
         if self.usepots and options.pot:
             outputoptions = {}
-            for (inputformat, templateformat), (outputformat, convertor) in self.outputoptions.iteritems():
+            for (inputformat, templateformat), (outputformat, convertor) in self.outputoptions.items():
                 inputformat = self.potifyformat(inputformat)
                 templateformat = self.potifyformat(templateformat)
                 outputformat = self.potifyformat(outputformat)
@@ -134,8 +141,9 @@ class ConvertOptionParser(optrecurse.RecursiveOptionParser, object):
             return self.outputoptions
 
     def setpotoption(self):
-        """Sets the ``-P``/``--pot`` option depending on input/output
-        formats etc."""
+        """Sets the ``-P``/``--pot`` option depending on input/output formats
+        etc.
+        """
         if self.usepots:
             potoption = optparse.Option(
                 "-P", "--pot",
@@ -154,18 +162,19 @@ class ConvertOptionParser(optrecurse.RecursiveOptionParser, object):
         self.define_option(timestampopt)
 
     def verifyoptions(self, options):
-        """Verifies that the options are valid (required options are
-        present, etc)."""
+        """Verifies that the options are valid (required options are present,
+        etc).
+        """
         pass
 
     def run(self, argv=None):
         """Parses the command line options and runs the conversion."""
         (options, args) = self.parse_args(argv)
-        options.inputformats = self.filterinputformats(options)
-        options.outputoptions = self.filteroutputoptions(options)
+        self.inputformats = self.filterinputformats(options)
+        self.outputoptions = self.filteroutputoptions(options)
         try:
             self.verifyoptions(options)
-        except Exception, e:
+        except Exception as e:
             self.error(str(e))
         self.recursiveprocess(options)
 
@@ -174,10 +183,10 @@ class ConvertOptionParser(optrecurse.RecursiveOptionParser, object):
         if options.timestamp and _output_is_newer(fullinputpath, fulloutputpath):
             return False
 
-        return super(ConvertOptionParser,
-                    self).processfile(fileprocessor, options,
-                                      fullinputpath, fulloutputpath,
-                                      fulltemplatepath)
+        return super().processfile(
+            fileprocessor, options, fullinputpath, fulloutputpath,
+            fulltemplatepath)
+
 
 def copyinput(inputfile, outputfile, templatefile, **kwargs):
     """Copies the input file to the output file."""
@@ -214,7 +223,8 @@ class Replacer:
     def searchreplacetemplate(self, inputfile, outputfile,
                               templatefile, **kwargs):
         """Copies the template file to the output file, searching and
-        replacing."""
+        replacing.
+        """
         outputfile.write(self.doreplace(templatefile.read()))
         return True
 
@@ -240,8 +250,9 @@ class ArchiveConvertOptionParser(ConvertOptionParser):
     ``archiveformats`` maps extension to class. If the extension doesn't
     matter, it can be None.
 
-    If the extension is only valid for input/output/template, it can be
-    given as ``(extension, filepurpose)``."""
+    If the extension is only valid for input/output/template, it can be given
+    as ``(extension, filepurpose)``.
+    """
 
     def __init__(self, formats, usetemplates=False, usepots=False,
                  description=None, archiveformats=None):
@@ -250,23 +261,17 @@ class ArchiveConvertOptionParser(ConvertOptionParser):
         else:
             self.archiveformats = archiveformats
         self.archiveoptions = {}
-        ConvertOptionParser.__init__(self, formats, usetemplates, usepots,
-                                     description=description)
-
-    def setarchiveoptions(self, **kwargs):
-        """Allows setting options that will always be passed to openarchive."""
-        self.archiveoptions = kwargs
+        super().__init__(formats, usetemplates, usepots, description=description)
 
     def isrecursive(self, fileoption, filepurpose='input'):
         """Checks if **fileoption** is a recursive file."""
         if self.isarchive(fileoption, filepurpose):
             return True
-        return super(ArchiveConvertOptionParser, self).isrecursive(fileoption,
-                                                                   filepurpose)
+        return super().isrecursive(fileoption, filepurpose)
 
     def isarchive(self, fileoption, filepurpose='input'):
         """Returns whether the file option is an archive file."""
-        if not isinstance(fileoption, (str, unicode)):
+        if not isinstance(fileoption, str):
             return False
         mustexist = (filepurpose != 'output')
         if mustexist and not os.path.isfile(fileoption):
@@ -303,13 +308,14 @@ class ArchiveConvertOptionParser(ConvertOptionParser):
         return archiveclass(archivefilename, **archiveoptions)
 
     def recurseinputfiles(self, options):
-        """Recurse through archive file / directories and return files
-        to be converted."""
+        """Recurse through archive file / directories and return files to be
+        converted.
+        """
         if self.isarchive(options.input, 'input'):
             options.inputarchive = self.openarchive(options.input, 'input')
             return self.recursearchivefiles(options)
         else:
-            return super(ArchiveConvertOptionParser, self).recurseinputfiles(options)
+            return super().recurseinputfiles(options)
 
     def recursearchivefiles(self, options):
         """Recurse through archive files and convert files."""
@@ -318,7 +324,7 @@ class ArchiveConvertOptionParser(ConvertOptionParser):
             if self.isexcluded(options, inputpath):
                 continue
             top, name = os.path.split(inputpath)
-            if not self.isvalidinputname(options, name):
+            if not self.isvalidinputname(name):
                 continue
             inputfiles.append(inputpath)
         return inputfiles
@@ -328,7 +334,7 @@ class ArchiveConvertOptionParser(ConvertOptionParser):
         if self.isarchive(options.input, 'input'):
             return options.inputarchive.openinputfile(fullinputpath)
         else:
-            return super(ArchiveConvertOptionParser, self).openinputfile(options, fullinputpath)
+            return super().openinputfile(options, fullinputpath)
 
     def getfullinputpath(self, options, inputpath):
         """Gets the absolute path to an input file."""
@@ -343,11 +349,11 @@ class ArchiveConvertOptionParser(ConvertOptionParser):
             if (options.recursivetemplate and
                 self.isarchive(options.template, 'template')):
                 # TODO: deal with different names in input/template archives
-                if fulltemplatepath in options.templatearchive:
-                    return options.templatearchive.openinputfile(fulltemplatepath)
+                if fulltemplatepath in self.templatearchive:
+                    return self.templatearchive.openinputfile(fulltemplatepath)
                 else:
                     self.warning("missing template file %s" % fulltemplatepath)
-        return super(ArchiveConvertOptionParser, self).opentemplatefile(options, fulltemplatepath)
+        return super().opentemplatefile(options, fulltemplatepath)
 
     def getfulltemplatepath(self, options, templatepath):
         """Gets the absolute path to a template file."""
@@ -366,8 +372,8 @@ class ArchiveConvertOptionParser(ConvertOptionParser):
         if templatepath is not None:
             if self.isarchive(options.template, 'template'):
                 # TODO: deal with different names in input/template archives
-                return templatepath in options.templatearchive
-        return super(ArchiveConvertOptionParser, self).templateexists(options, templatepath)
+                return templatepath in self.templatearchive
+        return super().templateexists(options, templatepath)
 
     def getfulloutputpath(self, options, outputpath):
         """Gets the absolute path to an output file."""
@@ -380,49 +386,41 @@ class ArchiveConvertOptionParser(ConvertOptionParser):
 
     def checkoutputsubdir(self, options, subdir):
         """Checks to see if subdir under ``options.output`` needs to be
-        created, creates if neccessary."""
+        created, creates if neccessary.
+        """
         if not self.isarchive(options.output, 'output'):
-            super(ArchiveConvertOptionParser, self).checkoutputsubdir(options, subdir)
+            super().checkoutputsubdir(options, subdir)
 
     def openoutputfile(self, options, fulloutputpath):
         """Opens the output file."""
         if self.isarchive(options.output, 'output'):
-            outputstream = options.outputarchive.openoutputfile(fulloutputpath)
+            outputstream = self.outputarchive.openoutputfile(fulloutputpath)
             if outputstream is None:
-                self.warning("Could not find where to put %s in output archive; writing to tmp" % fulloutputpath)
-                return StringIO()
+                self.warning("Could not find where to put %s in output "
+                             "archive; writing to tmp" % fulloutputpath)
+                return BytesIO()
             return outputstream
         else:
-            return super(ArchiveConvertOptionParser, self).openoutputfile(options, fulloutputpath)
-
-    def inittemplatearchive(self, options):
-        """Opens the ``templatearchive`` if not already open."""
-        if not self.usetemplates:
-            return
-        if (options.template and
-            self.isarchive(options.template, 'template') and
-            not hasattr(options, "templatearchive")):
-            options.templatearchive = self.openarchive(options.template,
-                                                       'template')
-
-    def initoutputarchive(self, options):
-        """Creates an outputarchive if required."""
-        if options.output and self.isarchive(options.output, 'output'):
-            options.outputarchive = self.openarchive(options.output, 'output',
-                                                     mode="w")
+            return super().openoutputfile(options, fulloutputpath)
 
     def recursiveprocess(self, options):
         """Recurse through directories and convert files."""
         if hasattr(options, "multifilestyle"):
-            self.setarchiveoptions(multifilestyle=options.multifilestyle)
+            self.archiveoptions = {'multifilestyle': options.multifilestyle}
             for filetype in ("input", "output", "template"):
                 allowoption = "allowrecursive%s" % filetype
                 if (options.multifilestyle == "onefile" and
                     getattr(options, allowoption, True)):
                     setattr(options, allowoption, False)
-        self.inittemplatearchive(options)
-        self.initoutputarchive(options)
-        return super(ArchiveConvertOptionParser, self).recursiveprocess(options)
+
+        if (self.usetemplates and options.template and
+            self.isarchive(options.template, 'template')):
+            self.templatearchive = self.openarchive(options.template, 'template')
+
+        if options.output and self.isarchive(options.output, 'output'):
+            self.outputarchive = self.openarchive(options.output, 'output',
+                                                  mode="w")
+        return super().recursiveprocess(options)
 
     def processfile(self, fileprocessor, options, fullinputpath,
                     fulloutputpath, fulltemplatepath):
@@ -447,10 +445,10 @@ class ArchiveConvertOptionParser(ConvertOptionParser):
                     os.unlink(fulloutputpath)
                 return False
         else:
-            return super(ArchiveConvertOptionParser,
-                        self).processfile(fileprocessor, options,
-                                          fullinputpath, fulloutputpath,
-                                          fulltemplatepath)
+            return super().processfile(
+                fileprocessor, options, fullinputpath, fulloutputpath,
+                fulltemplatepath)
+
 
 def _output_is_newer(input_path, output_path):
     """Check if input_path was not modified since output_path was generated,
@@ -467,6 +465,7 @@ def _output_is_newer(input_path, output_path):
 
     return output_mtime > input_mtime
 
+
 def should_output_store(store, threshold):
     """Check if the percent of translated source words more than or equal to
     the given threshold.
@@ -477,16 +476,20 @@ def should_output_store(store, threshold):
 
     from translate.storage import statsdb
 
-    units = filter(lambda unit: unit.istranslatable(), store.units)
-    translated = filter(lambda unit: unit.istranslated(), units)
-    wordcounts = dict(map(lambda unit: (unit, statsdb.wordsinunit(unit)), units))
-    sourcewords = lambda elementlist: sum(map(lambda unit: wordcounts[unit][0], elementlist))
-    tranlated_count = sourcewords(translated)
+    units = list(filter(lambda unit: unit.istranslatable(), store.units))
+    translated = list(filter(lambda unit: unit.istranslated(), units))
+    wordcounts = dict(map(lambda unit: (unit.getid(), statsdb.wordsinunit(unit)), units))
+
+    def sourcewords(elementlist):
+        return sum(map(lambda unit: wordcounts[unit.getid()][0], elementlist))
+
+    translated_count = sourcewords(translated)
     total_count = sourcewords(units)
-    percent = tranlated_count * 100 / total_count
+    percent = translated_count * 100 / total_count
 
     return percent >= threshold
 
+
 def main(argv=None):
     parser = ArchiveConvertOptionParser({}, description=__doc__)
     parser.run(argv)
