@@ -1,5 +1,3 @@
-#!/usr/bin/env python
-# -*- coding: utf-8 -*-
 #
 # Copyright 2002-2009,2012 Zuza Software Foundation
 #
@@ -26,15 +24,15 @@
 
 import warnings
 
-from translate.storage import dtd
-from translate.storage import po
+from translate.convert import accesskey, convert
 from translate.misc import quote
-from translate.convert import accesskey
-from translate.convert import convert
+from translate.storage import dtd, po
 
 
 def dtdwarning(message, category, filename, lineno, line=None):
     return "Warning: %s\n" % message
+
+
 warnings.formatwarning = dtdwarning
 
 
@@ -58,6 +56,8 @@ def applytranslation(entity, dtdunit, inputunit, mixedentities):
                     label, unquotedstr = accesskey.extract(unquotedstr)
                     if not unquotedstr:
                         warnings.warn("Could not find accesskey for %s" % entity)
+                        # Use the source language accesskey
+                        label, unquotedstr = accesskey.extract(inputunit.source)
                     else:
                         original = dtdunit.source
                         # For the sake of diffs we keep the case of the
@@ -68,8 +68,7 @@ def applytranslation(entity, dtdunit, inputunit, mixedentities):
                                 unquotedstr = unquotedstr.upper()
                             elif original.islower():
                                 unquotedstr = unquotedstr.lower()
-    if len(unquotedstr) > 0:
-        dtdunit.source = dtd.removeinvalidamps(entity, unquotedstr)
+    dtdunit.source = unquotedstr
 
 
 class redtd:
@@ -89,10 +88,11 @@ class redtd:
     def handleinunit(self, inunit, includefuzzy):
         entities = inunit.getlocations()
         mixedentities = self.mixer.match_entities(entities)
+        self.dtdfile.require_index()
         for entity in entities:
-            if entity in self.dtdfile.index:
+            if entity in self.dtdfile.id_index:
                 # now we need to replace the definition of entity with msgstr
-                dtdunit = self.dtdfile.index[entity]  # find the dtd
+                dtdunit = self.dtdfile.id_index[entity]  # find the dtd
                 if inunit.istranslated() or not bool(inunit.source):
                     applytranslation(entity, dtdunit, inunit, mixedentities)
                 elif self.remove_untranslated and not (includefuzzy and inunit.isfuzzy()):
@@ -141,7 +141,7 @@ class po2dtd:
             unquoted = None
         else:
             unquoted = inputunit.source
-        dtdunit.source = dtd.removeinvalidamps(dtdunit.entity, unquoted)
+        dtdunit.source = unquoted
 
     def convertunit(self, inputunit):
         dtdunit = dtd.dtdunit()
@@ -173,7 +173,7 @@ def convertdtd(inputfile, outputfile, templatefile, includefuzzy=False,
     # identify them seems to be on their file path in the tree (based on code
     # in compare-locales).
     android_dtd = False
-    header_comment = u""
+    header_comment = ""
     input_header = inputstore.header()
     if input_header:
         header_comment = input_header.getnotes("developer")
@@ -188,7 +188,7 @@ def convertdtd(inputfile, outputfile, templatefile, includefuzzy=False,
         convertor = redtd(templatestore, android=android_dtd,
                           remove_untranslated=remove_untranslated)
     outputstore = convertor.convertstore(inputstore, includefuzzy)
-    outputfile.write(str(outputstore))
+    outputstore.serialize(outputfile)
     return 1
 
 
@@ -196,12 +196,9 @@ def main(argv=None):
     # handle command line options
     formats = {"po": ("dtd", convertdtd), ("po", "dtd"): ("dtd", convertdtd)}
     parser = convert.ConvertOptionParser(formats, usetemplates=True, description=__doc__)
-    parser.add_option("", "--removeuntranslated", dest="remove_untranslated",
-            default=False, action="store_true",
-            help="remove untranslated strings from output")
     parser.add_threshold_option()
     parser.add_fuzzy_option()
-    parser.passthrough.append("remove_untranslated")
+    parser.add_remove_untranslated_option()
     parser.run(argv)
 
 
