@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 #
 # Copyright 2006-2009 Zuza Software Foundation
 #
@@ -19,19 +18,17 @@
 
 """Base classes for storage interfaces."""
 
+from collections import OrderedDict
 import codecs
 import logging
-import six
-try:
-    import cPickle as pickle
-except ImportError:
-    import pickle
+import pickle
 from io import BytesIO
 
 from translate.misc.multistring import multistring
 from translate.storage.placeables import StringElem, parse as rich_parse
 from translate.storage.workflow import StateEnum as states
 
+
 # Simple BOM based encoding detection
 ENCODING_BOMS = (
     (codecs.BOM_UTF8, 'utf-8-sig'),
@@ -53,8 +50,7 @@ class ParseError(Exception):
         return repr(self.inner_exc)
 
 
-@six.python_2_unicode_compatible
-class TranslationUnit(object):
+class TranslationUnit:
     """Base class for translation units.
 
     Our concept of a *translation unit* is influenced heavily by `XLIFF
@@ -159,41 +155,54 @@ class TranslationUnit(object):
         >>> from translate.storage.placeables.interfaces import X
         >>> rich = [StringElem(['foo', X(id='xxx', sub=[' ']), 'bar'])]
         >>> TranslationUnit.rich_to_multistring(rich)
-        multistring(u'foo bar')
+        multistring('foo bar')
         """
-        return multistring([six.text_type(elem) for elem in elem_list])
+        return multistring([str(elem) for elem in elem_list])
 
     def multistring_to_rich(self, mulstring):
         """Convert a multistring to a list of "rich" string trees:
 
-        >>> target = multistring([u'foo', u'bar', u'baz'])
+        >>> target = multistring(['foo', 'bar', 'baz'])
         >>> TranslationUnit.multistring_to_rich(target)
-        [<StringElem([<StringElem([u'foo'])>])>,
-         <StringElem([<StringElem([u'bar'])>])>,
-         <StringElem([<StringElem([u'baz'])>])>]
+        [<StringElem([<StringElem(['foo'])>])>,
+         <StringElem([<StringElem(['bar'])>])>,
+         <StringElem([<StringElem(['baz'])>])>]
         """
         if isinstance(mulstring, multistring):
             return [rich_parse(s, self.rich_parsers) for s in mulstring.strings]
         return [rich_parse(mulstring, self.rich_parsers)]
 
-    def setsource(self, source):
+    @property
+    def source(self):
+        return self._source
+
+    @source.setter
+    def source(self, source):
         """Set the source string to the given value."""
         self._rich_source = None
         self._source = source
-    source = property(lambda self: self._source, setsource)
 
-    def settarget(self, target):
+    @property
+    def target(self):
+        return self._target
+
+    @target.setter
+    def target(self, target):
         """Set the target string to the given value."""
         self._rich_target = None
         self._target = target
-    target = property(lambda self: self._target, settarget)
 
-    def _get_rich_source(self):
+    @property
+    def rich_source(self):
+        """
+        .. seealso:: :meth:`.rich_to_multistring`, :meth:`multistring_to_rich`
+        """
         if self._rich_source is None:
             self._rich_source = self.multistring_to_rich(self.source)
         return self._rich_source
 
-    def _set_rich_source(self, value):
+    @rich_source.setter
+    def rich_source(self, value):
         if not hasattr(value, '__iter__'):
             raise ValueError('value must be iterable')
         if len(value) < 1:
@@ -204,17 +213,18 @@ class TranslationUnit(object):
         multi = self.rich_to_multistring(value)
         if self.source != multi:
             self.source = multi
-    rich_source = property(_get_rich_source, _set_rich_source)
-    """
-    .. seealso:: :meth:`.rich_to_multistring`, :meth:`multistring_to_rich`
-    """
 
-    def _get_rich_target(self):
+    @property
+    def rich_target(self):
+        """
+        .. seealso:: :meth:`.rich_to_multistring`, :meth:`.multistring_to_rich`
+        """
         if self._rich_target is None:
             self._rich_target = self.multistring_to_rich(self.target)
         return self._rich_target
 
-    def _set_rich_target(self, value):
+    @rich_target.setter
+    def rich_target(self, value):
         if not hasattr(value, '__iter__'):
             raise ValueError('value must be iterable')
         if len(value) < 1:
@@ -223,10 +233,6 @@ class TranslationUnit(object):
             raise ValueError('value[0] must be of type StringElem.')
         self._rich_target = list(value)
         self.target = self.rich_to_multistring(value)
-    rich_target = property(_get_rich_target, _set_rich_target)
-    """
-    .. seealso:: :meth:`.rich_to_multistring`, :meth:`.multistring_to_rich`
-    """
 
     def gettargetlen(self):
         """Returns the length of the target string.
@@ -334,9 +340,9 @@ class TranslationUnit(object):
         else:
             self.notes = text
 
-    def removenotes(self):
+    def removenotes(self, origin=None):
         """Remove all the translator's notes."""
-        self.notes = u''
+        self.notes = ''
 
     def adderror(self, errorname, errortext):
         """Adds an error message to this unit.
@@ -464,7 +470,7 @@ class TranslationUnit(object):
     def get_state_id(self, n=None):
         if n is None:
             n = self.get_state_n()
-        for state_id, state_range in six.iteritems(self.STATE):
+        for state_id, state_range in self.STATE.items():
             if state_range[0] <= n < state_range[1]:
                 return state_id
         if self.STATE:
@@ -488,7 +494,7 @@ class TranslationUnit(object):
         pass
 
 
-class TranslationStore(object):
+class TranslationStore:
     """Base class for stores for multiple translation units of type UnitClass.
     """
 
@@ -516,6 +522,9 @@ class TranslationStore(object):
         if unitclass:
             self.UnitClass = unitclass
         self._encoding = encoding
+        self.locationindex = {}
+        self.sourceindex = {}
+        self.id_index = {}
 
     @property
     def encoding(self):
@@ -577,6 +586,18 @@ class TranslationStore(object):
         unit._store = self
         self.units.append(unit)
 
+    def removeunit(self, unit):
+        """Remove the given unit to the object's list of units.
+
+        This method should always be used rather than trying to modify the
+        list manually.
+
+        :type unit: :class:`TranslationUnit`
+        :param unit: The unit that will be added.
+        """
+        self.units.remove(unit)
+        self.remove_unit_from_index(unit)
+
     def addsourceunit(self, source):
         """Add and returns a new unit with the given source string.
 
@@ -596,13 +617,9 @@ class TranslationStore(object):
 
         :rtype: :class:`TranslationUnit` or None
         """
-        if len(getattr(self, "sourceindex", [])):
-            if source in self.sourceindex:
-                return self.sourceindex[source][0]
-        else:
-            for unit in self.units:
-                if unit.source == source:
-                    return unit
+        self.require_index()
+        if source in self.sourceindex:
+            return self.sourceindex[source][0]
         return None
 
     def findunits(self, source):
@@ -610,17 +627,9 @@ class TranslationStore(object):
 
         :rtype: :class:`TranslationUnit` or None
         """
-        if len(getattr(self, "sourceindex", [])):
-            if source in self.sourceindex:
-                return self.sourceindex[source]
-        else:
-            #FIXME: maybe we should generate index here instead since
-            #we'll scan all units anyway
-            result = []
-            for unit in self.units:
-                if unit.source == source:
-                    result.append(unit)
-            return result
+        self.require_index()
+        if source in self.sourceindex:
+            return self.sourceindex[source]
         return None
 
     def translate(self, source):
@@ -637,20 +646,17 @@ class TranslationStore(object):
     def remove_unit_from_index(self, unit):
         """Remove a unit from source and locaton indexes"""
 
-        def remove_unit(source):
+        def remove_source(source):
             if source in self.sourceindex:
-                try:
-                    self.sourceindex[source].remove(unit)
-                    if len(self.sourceindex[source]) == 0:
-                        del(self.sourceindex[source])
-                except ValueError:
-                    pass
+                self.sourceindex[source].remove(unit)
+                if len(self.sourceindex[source]) == 0:
+                    del(self.sourceindex[source])
 
         if unit.hasplural():
             for source in unit.source.strings:
-                remove_unit(source)
+                remove_source(source)
         else:
-            remove_unit(unit.source)
+            remove_source(unit.source)
 
         for location in unit.getlocations():
             if (location in self.locationindex and
@@ -695,7 +701,7 @@ class TranslationStore(object):
 
     def require_index(self):
         """make sure source index exists"""
-        if not hasattr(self, "id_index"):
+        if not self.id_index:
             self.makeindex()
 
     def getids(self, filename=None):
@@ -709,13 +715,6 @@ class TranslationStore(object):
         odict['fileobj'] = None
         return odict
 
-    def __str__(self):
-        # This allows the old str(store) method for serialization to be kept
-        # for compatibility purpose.
-        if six.PY2:
-            return self.__bytes__()
-        return super(TranslationStore, self).__str__()
-
     def __bytes__(self):
         out = BytesIO()
         self.serialize(out)
@@ -753,11 +752,10 @@ class TranslationStore(object):
     def parsestring(cls, storestring):
         """Convert the string representation back to an object."""
         newstore = cls()
-        if storestring:
-            if isinstance(storestring, six.text_type):
-                # parse() is expecting bytes
-                storestring = storestring.encode(cls.default_encoding)
-            newstore.parse(storestring)
+        if isinstance(storestring, str):
+            # parse() is expecting bytes
+            storestring = storestring.encode(cls.default_encoding)
+        newstore.parse(storestring)
         return newstore
 
     def fallback_detection(self, text):
@@ -812,18 +810,18 @@ class TranslationStore(object):
             if encoding == self.encoding and suffix == 'sig':
                 encodings.append(detected_encoding['encoding'])
             elif detected_encoding['encoding'] != self.encoding:
-                logging.warn("trying to parse %s with encoding: %s but "
-                             "detected encoding is %s (confidence: %s)",
-                             self.filename, self.encoding,
-                             detected_encoding['encoding'],
-                             detected_encoding['confidence'])
+                logging.warning("trying to parse %s with encoding: %s but "
+                                "detected encoding is %s (confidence: %s)",
+                                self.filename, self.encoding,
+                                detected_encoding['encoding'],
+                                detected_encoding['confidence'])
             encodings.append(self.encoding)
         else:
             encodings.append(self.encoding)
 
         for encoding in encodings:
             try:
-                r_text = six.text_type(text, encoding)
+                r_text = str(text, encoding)
                 r_encoding = encoding
                 break
             except UnicodeDecodeError:
@@ -839,7 +837,7 @@ class TranslationStore(object):
 
     def savefile(self, storefile):
         """Write the string representation to the given file (or filename)."""
-        if isinstance(storefile, six.string_types):
+        if isinstance(storefile, str):
             storefile = open(storefile, 'wb')
         self.fileobj = storefile
         self._assignname()
@@ -867,16 +865,17 @@ class TranslationStore(object):
         """Reads the given file (or opens the given filename) and parses back
         to an object.
         """
-        if isinstance(storefile, six.string_types):
+        if isinstance(storefile, str):
             storefile = open(storefile, 'rb')
         mode = getattr(storefile, "mode", 'rb')
         #For some reason GzipFile returns 1, so we have to test for that here
         if mode == 1 or "r" in mode:
             storestring = storefile.read()
             storefile.close()
+            newstore = cls.parsestring(storestring)
         else:
             storestring = ""
-        newstore = cls.parsestring(storestring)
+            newstore = cls()
         newstore.fileobj = storefile
         newstore._assignname()
         return newstore
@@ -889,3 +888,109 @@ class TranslationStore(object):
         :rtype: string
         """
         return "id"
+
+
+class UnitId:
+    KEY_SEPARATOR = "."
+    INDEX_SEPARATOR = ""
+
+    def __init__(self, parts):
+        self.parts = parts
+
+    def __str__(self):
+        def fmt(element, key):
+            if element == 'key':
+                return '{}{}'.format(self.KEY_SEPARATOR, key)
+            elif element == 'index':
+                return '{}[{}]'.format(self.INDEX_SEPARATOR, key)
+            else:
+                raise ValueError('Unsupported element: {}'.format(element))
+        return ''.join([fmt(*part) for part in self.parts])
+
+    def __add__(self, other):
+        if not isinstance(other, list):
+            raise ValueError('Not supported type for add: {}'.format(type(other)))
+        return self.__class__(self.parts + other)
+
+    @classmethod
+    def from_string(cls, text):
+        result = []
+        # Strip possible leading separator
+        if text.startswith(cls.KEY_SEPARATOR):
+            text = text[len(cls.KEY_SEPARATOR):]
+        for item in text.split(cls.KEY_SEPARATOR):
+            if '[' in item and item[-1] == ']':
+                item, pos = item[:-1].split('[')
+                if cls.INDEX_SEPARATOR and item:
+                    result.append(("key", item))
+                result.append(("index", int(pos)))
+            else:
+                result.append(("key", item))
+        return cls(result)
+
+
+class DictUnit(TranslationUnit):
+    IdClass = UnitId
+    DefaultDict = OrderedDict
+
+    def __init__(self, source=None):
+        super().__init__(source)
+        self._unitid = None
+
+    def storevalue(self, output, value, override_key=None, unset=False):
+        target = output
+        if self._unitid is None:
+            self._unitid = self.IdClass.from_string(self._id)
+        parts = self._unitid.parts
+        for pos, part in enumerate(parts[:-1]):
+            element, key = part
+            default = [] if parts[pos + 1][0] == 'index' else self.DefaultDict()
+            if element == 'index':
+                if len(target) <= key and not unset:
+                    target.append(default)
+            elif element == 'key':
+                if key not in target or isinstance(target[key], str):
+                    target[key] = default
+            else:
+                raise ValueError('Unsupported element: {}'.format(element))
+            target = target[key]
+        if override_key:
+            element, key = 'key', override_key
+        else:
+            element, key = parts[-1]
+        if element == 'key':
+            if unset:
+                del target[key]
+            else:
+                target[key] = value
+        elif element == 'index':
+            if len(target) <= key:
+                if not unset:
+                    target.append(value)
+            else:
+                if unset:
+                    del target[key]
+                else:
+                    target[key] = value
+        else:
+            raise ValueError('Unsupported element: {}'.format(element))
+
+    def storevalues(self, output):
+        self.storevalue(output, self.value)
+
+    def getvalue(self):
+        """Returns dictionary for serialization."""
+        result = {}
+        self.storevalues(result)
+        return result
+
+    def set_unitid(self, unitid):
+        self.setid(str(unitid))
+        self._unitid = unitid
+
+
+class DictStore(TranslationStore):
+
+    def serialize_units(self, output):
+        for unit in self.unit_iter():
+            unit.storevalues(output)
